import pytest

from fuel_efficency.entities.down_hill import DownHill
from fuel_efficency.entities.node import Node
from fuel_efficency.entities.plateau import Plateau
from fuel_efficency.entities.up_hill import UpHill
from fuel_efficency.entities.valley import Valley


@pytest.mark.parametrize("node", [Valley, UpHill, DownHill, Plateau])
def test_eq_error_raised(node:Node):
    valley = node()
    other = int()

    with pytest.raises(NotImplementedError) as excinfo:
        valley == other
    assert "Missing `position` or `weight` attribute" in str(excinfo.value)

@pytest.mark.parametrize("node", [Valley, UpHill, DownHill, Plateau])
def test_lt_error_raised(node:Node):
    valley = node()
    other = int()

    with pytest.raises(NotImplementedError) as excinfo:
        valley < other
    assert "Missing `weight` attribute" in str(excinfo.value)



import sys
from typing import TypeVar

import pytest

from fuel_efficency.entities.down_hill import DownHill
from fuel_efficency.entities.node import Node
from fuel_efficency.entities.plateau import Plateau
from fuel_efficency.entities.position import Position
from fuel_efficency.entities.up_hill import UpHill
from fuel_efficency.entities.valley import Valley


@pytest.mark.parametrize("expected_x, expected_y", [
    (0, 0),
    (1, 1),
    (2, 2),
    (3, 3),
    (4, 4)
])
def test_position(expected_x: int, expected_y: int) -> None:
    """
    Test the Position class.

    Args:
        expected_x (int): The expected x coordinate.
        expected_y (int): The expected y coordinate.

    Returns:
        None: This function should not return anything.
    """
    position = Position(x=expected_x, y=expected_y)
    assert isinstance(position.x, int), "Position x is not an instance of int."
    assert isinstance(position.y, int), "Position y is not an instance of int."
    assert position.x == expected_x, "Position x is not as expected."
    assert position.y == expected_y, "Position y is not as expected."

@pytest.mark.parametrize("node_protocol, expected_weight", [
    (Valley, float(1)),
    (Plateau, float(1)),
    (UpHill, float(2)),
    (DownHill, float(0.5)),
    (Node, float("inf"))
])
def test_node_weights(node_protocol:Node, expected_weight:float) -> None:
    """
    Test that the weight property of each node class is correct.

    Args:
        node_protocol (Node): The node class to test.
        expected_weight (float): The expected weight of the node class.

    Returns:
        None: This function should not return anything.
    """
    if node_protocol is not Node:
        node_protocol:Node = node_protocol()
        assert node_protocol.weight == expected_weight, f"Incorrect weight for {node_protocol.__name__}"
    else:
        with pytest.raises(TypeError):
            node_protocol()

@pytest.mark.parametrize("node_protocol", [
    (Valley),
    (Plateau),
    (UpHill),
    (DownHill)
])
def test_node_equality(node_protocol:Node) -> None:
    """
    Test that the node protocol is equal to itself.

    Args:
        node_protocol (Node): The node protocol to test.

    Returns:
        None: This function should not return anything.
    """
    node_protocol:Node = node_protocol()
    assert node_protocol == node_protocol, f"{node_protocol.__name__} is not equal to itself."
    assert hash(node_protocol) == hash(node_protocol), f"{node_protocol.__name__} hash is not equal to itself."
    MockNode = TypeVar("MockNode")
    mock_node:MockNode = Valley(weight=sys.maxsize)
    assert node_protocol < mock_node

@pytest.mark.parametrize("node_protocol", [
    (Valley),
    (Plateau),
    (UpHill),
    (DownHill)
])
def test_node_hash(node_protocol:Node) -> None:
    """
    Test that the node protocol hash is equal to itself.

    Args:
        node_protocol (Node): The node protocol to test.

    Returns:
        None: This function should not return anything.
    """
    node_protocol:Node = node_protocol()
    assert hash(node_protocol) == hash(node_protocol), f"{node_protocol.__name__} hash is not equal to itself."

@pytest.mark.parametrize("node_protocol", [
(Valley),
(Plateau),
(UpHill),
(DownHill)
])
def test_node_less_than(node_protocol:Node) -> None:
    """
    Test that the node protocol is less than a mock node.

    Args:
        node_protocol (Node): The node protocol to test.

    Returns:
        None: This function should not return anything.
    """
    node_protocol:Node = node_protocol()
    MockNode = TypeVar("MockNode")
    mock_node:MockNode = Valley(weight=sys.maxsize)
    assert node_protocol < mock_node

@pytest.mark.parametrize("node_protocol, expected_position", [
    (Valley, Position(x=0, y=0)),
    (Plateau, Position(x=0, y=0)),
    (UpHill, Position(x=0, y=0)),
    (DownHill, Position(x=0, y=0)),
    (Node, Position())
])
def test_node_positions(node_protocol:Node, expected_position:Position) -> None:
    """
    Test that the position property of each node class is correct.

    Args:
        node_protocol (Node): The node class to test.
        expected_position (Position): The expected position of the node class.

    Returns:
        None: This function should not return anything.
    """
    if node_protocol is not Node:
        node_protocol = node_protocol(position=expected_position)
        assert node_protocol.position == expected_position, f"Incorrect position for {node_protocol.__name__}"
    else:
        with pytest.raises(TypeError):
            node_protocol()

def test_position_add_success() -> None:
    """
    Test that the position class can be added to another position class.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    position1 = Position(x=10, y=20)
    position2 = Position(x=5, y=3)

    result = position1 + position2
    assert result.x == 15 and result.y == 23, "Addition should correctly add Position coordinates"

def test_position_add_invalid_type() -> None:
    """
    Test that the position class can be added to another position class.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    position = Position(x=10, y=20)
    other = "not a position object"

    with pytest.raises(NotImplementedError) as excinfo:
        _ = position + other
    assert f"Cannot add Position and {type(other)}" in str(excinfo.value)

def test_position_sub_success() -> None:
    """
    Test that the position class can be subtracted from another position class.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    position1 = Position(x=10, y=20)
    position2 = Position(x=5, y=3)

    result = position1 - position2
    assert result.x == 5 and result.y == 17, "Subtraction should correctly subtract Position coordinates"

def test_position_sub_invalid_type() -> None:
    """
    Test that the position class can be subtracted from another position class.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    position = Position(x=10, y=20)
    other = "not a position object"

    with pytest.raises(NotImplementedError) as excinfo:
        _ = position - other
    assert f"Cannot subtract Position and {type(other)}" in str(excinfo.value)



from dataclasses import dataclass, field
from functools import total_ordering

from fuel_efficency.entities.node import Node
from fuel_efficency.entities.position import Position


@total_ordering
@dataclass(slots=True)
class Plateau:
    weight: float = float(1)
    position: Position = field(default_factory=Position)

    def __eq__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError("Missing `position` or `weight` attribute")
        if self is other:
            return True
        return self.weight == other.weight and self.position == other.position

    def __gt__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError("Missing `weight` attribute")
        return self.weight > other.weight

    def __hash__(self):
        return id(self)



import sys
import heapq
from typing import List

from fuel_efficency.algorithms.path_finding import PathfindingStrategy
from fuel_efficency.entities.node import Node
from fuel_efficency.entities.position import Position


class AStarStrategy(PathfindingStrategy):

    allowed_directions = [Position(-1, 0), Position(0, -1), Position(0, 1), Position(1, 0)]

    @staticmethod
    def find_path(grid:List[List[Node]], start:Node, end:Node):
        priority_queue = [(0, start)]
        visited_nodes = set()
        cum_distance = {node: 0 if node is start else sys.maxsize for row in grid for node in row}
        previous_node = {node: None for row in grid for node in row}

        while priority_queue:
            current_cum_distance, current_node = heapq.heappop(priority_queue)
            visited_nodes.add(current_node)

            if current_node == end:
                path = []
                while current_node:
                    path.insert(0, current_node)
                    current_node = previous_node[current_node]
                return path[1:]

            for neighbor in AStarStrategy.get_neighbors(grid, current_node):
                if neighbor in visited_nodes:
                    continue

                inc_distance = AStarStrategy.calculate_distance(current_node, neighbor)
                heuristic_distance = AStarStrategy.get_heuristic_distance(neighbor, end)
                new_cum_distance = current_cum_distance + inc_distance + heuristic_distance
                if new_cum_distance < cum_distance[neighbor]:
                    cum_distance[neighbor] = new_cum_distance
                    previous_node[neighbor] = current_node
                    heapq.heappush(priority_queue, (new_cum_distance, neighbor))

    @staticmethod
    def get_neighbors(grid:List[List[Node]], node:Node) -> List[Node]:
        row_qty = len(grid)
        col_qty = len(grid[0])

        neighbors = []
        for offset in AStarStrategy.allowed_directions:
            neighbor_position = node.position - offset
            if 0 <= neighbor_position.x < row_qty and 0 <= neighbor_position.y < col_qty:
                neightbor = grid[neighbor_position.x][neighbor_position.y]
                neighbors.append(neightbor)

        return neighbors

    @staticmethod
    def calculate_distance(_:Node, node2: Node) -> float:
        # here we are assuming only orthogonal movements are valid as defined in `allowed_directions`
        # so no need for the sqrt(2) for diagonal movements
        return node2.weight

    @staticmethod
    def get_heuristic_distance(node1:Node, node2: Node):
        # here we are using the traditional manhattan distance
        offset_x = abs(node1.position.x - node2.position.x)
        offset_y = abs(node1.position.y - node2.position.y)
        return offset_x + offset_y



from abc import ABC, abstractmethod
from typing import List

from fuel_efficency.entities.node import Node


class PathfindingStrategy(ABC):

    @abstractmethod
    def find_path(grid:List[List[Node]], start:Node, end:Node):
        pass # pragma: no cover

    @abstractmethod
    def get_neighbors(grid:List[List[Node]], node:Node) -> List[Node]:
        pass # pragma: no cover

    @abstractmethod
    def calculate_distance(node1:Node, node2: Node) -> float:
        pass # pragma: no cover



from dataclasses import dataclass, field
from typing import Protocol, runtime_checkable

from fuel_efficency.entities.position import Position


@runtime_checkable
@dataclass(slots=True)
class Node(Protocol):
    weight: float
    position: Position =  field(default_factory=Position)



import math
from typing import List

import pytest

from fuel_efficency.algorithms.a_star import AStarStrategy
from fuel_efficency.algorithms.context import Context
from fuel_efficency.algorithms.dijkstra import DijkstraStrategy
from fuel_efficency.algorithms.path_finding import PathfindingStrategy
from fuel_efficency.entities.node import Node
from fuel_efficency.entities.position import Position
from fuel_efficency.entities.valley import Valley
from fuel_efficency.entities.up_hill import UpHill


def create_3_by_3_flat_terrain_grid():
    return [[Valley(position=Position(x, y)) for y in range(3)] for x in range(3)]


test_data = [
    (Valley(position=Position(1, 1)), 8),  # Center
    (Valley(position=Position(0, 0)), 3),  # Top left corner
    (Valley(position=Position(0, 1)), 5),  # Top edge
    (Valley(position=Position(0, 2)), 3),  # Top right corner
    (Valley(position=Position(1, 0)), 5),  # Left edge
    (Valley(position=Position(1, 2)), 5),  # Right edge
    (Valley(position=Position(2, 0)), 3),  # Bottom left corner
    (Valley(position=Position(2, 1)), 5),  # Bottom edge
    (Valley(position=Position(2, 2)), 3),  # Bottom right corner
]

@pytest.mark.parametrize("start_node, expected_neighbors_count", test_data)
def test_get_neighbors(start_node:Node, expected_neighbors_count:int):
    grid = create_3_by_3_flat_terrain_grid()
    grid_width = len(grid)
    grid_height = len(grid[0])
    neighbors = DijkstraStrategy.get_neighbors(grid, start_node)
    assert len(neighbors) == expected_neighbors_count
    assert all(0 <= neighbor.position.x < grid_width and 0 <= neighbor.position.y < grid_height for neighbor in neighbors)


distance_test_data = [
    (Valley(position=Position(0, 0)), Valley(position=Position(1, 1)), math.sqrt(2)),  # Diagonal distance
    (Valley(position=Position(0, 0)), Valley(position=Position(0, 1)), 1),  # Vertical distance
    (Valley(position=Position(0, 0)), Valley(position=Position(1, 0)), 1),  # Horizontal distance
]

@pytest.mark.parametrize("node1, node2, expected_distance", distance_test_data)
def test_calculate_distance(node1: Node, node2: Node, expected_distance: float):
    calculated_distance = DijkstraStrategy.calculate_distance(node1, node2)
    assert math.isclose(calculated_distance, expected_distance, rel_tol=1e-9)

def test_dijkstra_find_path():
    grid = create_3_by_3_flat_terrain_grid()
    start = grid[0][0]  # Top-left corner
    end = grid[2][2]  # Bottom-right corner

    # Expected path: straight line from top-left to bottom-right ommitting the start position as there's no weight
    expected_path = [
        grid[x][x] for x in range(1, 3)
    ]

    path = DijkstraStrategy.find_path(grid, start, end)
    assert path == expected_path

def test_astar_find_path():
    grid = create_3_by_3_flat_terrain_grid()
    start = grid[0][0]  # Top-left corner
    end = grid[2][2]  # Bottom-right corner

    # Expected path: straight line from top-left to bottom-right ommitting the start position as there's no weight
    positions = [Position(0, 1), Position(0, 2), Position(1, 2), Position(2, 2)]

    expected_path = [Valley(position=position) for position in positions]

    path = AStarStrategy.find_path(grid, start, end)

    assert path == expected_path

grid = create_3_by_3_flat_terrain_grid()
start_node = grid[0][0]  # Top-left corner
end_node = grid[2][2]  # Bottom-right corner

test_cases = [
    (AStarStrategy(), grid, start_node, end_node),
    (DijkstraStrategy(), grid, start_node, end_node)
]

from unittest.mock import MagicMock


@pytest.mark.parametrize("strategy, grid, start, end", test_cases)
def test_context_run(strategy:PathfindingStrategy, grid:List[List[Node]], start:Node, end:Node):
    strategy.find_path = MagicMock(spec=PathfindingStrategy.find_path)
    context:Context = Context(_strategy=strategy, _grid=grid, _start=start, _end=end)
    _ = context.run()
    strategy.find_path.assert_called_once_with(grid, start, end)
    assert context.strategy == strategy
    assert context.grid == grid
    assert context.start == start
    assert context.end == end

def test_grid_setter_invalid_type() -> None:
    """
    Test that an error is raised when an invalid type is set for the grid attribute.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    context = Context(_strategy=AStarStrategy(), _grid=[], _start=Valley(), _end=Valley())
    invalid_grid = "not a list"

    with pytest.raises(TypeError) as excinfo:
        context.grid = invalid_grid
    assert "Grid must be a list" in str(excinfo.value)

def test_grid_setter_invalid_row_type() -> None:
    """
    Test that an error is raised when the grid attribute is not a list of lists.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    context = Context(_strategy=AStarStrategy(), _grid=[], _start=Valley(), _end=Valley())
    invalid_grid = [10, 20]  # Not a list of lists

    with pytest.raises(TypeError) as excinfo:
        context.grid = invalid_grid
    assert "Grid must be a list of lists" in str(excinfo.value)

def test_strategy_setter_invalid_type() -> None:
    """
    Test that an error is raised when an invalid type is set for the strategy attribute.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    context = Context(_strategy=AStarStrategy(), _grid=[], _start=Valley(), _end=Valley())
    invalid_strategy = "not a PathfindingStrategy object"

    with pytest.raises(TypeError) as excinfo:
        context.strategy = invalid_strategy
    assert "Strategy must be an instance of PathfindingStrategy" in str(excinfo.value)

def test_run_method_missing_find_path() -> None:
    """
    Test that an error is raised when the strategy does not implement the find_path method.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """
    context = Context(_strategy=object(), _grid=[], _start=Valley(), _end=Valley())

    with pytest.raises(NotImplementedError) as excinfo:
        context.run()
    assert "Strategy must implement the find_path method" in str(excinfo.value)


def test_context_run_success():
    """
    Test that the Context class successfully runs the pathfinding strategy.

    Args:
        None

    Returns:
        None: This function should not return anything.
    """

    class MockPathfindingStrategy(PathfindingStrategy):
        def find_path(self, grid, start:Node, end:Node):
            # Implement a simple path finding logic for testing,
            # or return a predefined path
            return [Valley(position=Position(x, y)) for x in range(start.position.x, end.position.x + 1) for y in range(start.position.y, end.position.y + 1)]
        def calculate_distance(self, node1, node2):
            pass
        def get_neighbors(self, grid, node):
            pass

    grid = [[Valley() for _ in range(3)] for _ in range(3)]
    start = Valley(position=Position(0, 0))  # Top-left corner
    end = Valley(position=Position(2, 2))    # Bottom-right corner
    strategy = MockPathfindingStrategy()

    context:Context = Context()
    context.grid = grid
    context.start = start
    context.end = end
    context.strategy = strategy

    # Define expected path based on the MockPathfindingStrategy's logic
    expected_path = [Valley(position=Position(x, y)) for x in range(start.position.x, end.position.x + 1) for y in range(start.position.y, end.position.y + 1)]

    path = context.run()

    assert path == expected_path, "The path should match the expected path"

def test_astar_lower_border_path():
    grid = [
        [Valley(position=Position(0, 0)), UpHill(position=Position(0, 1)), UpHill(position=Position(0, 2))],
        [Valley(position=Position(1, 0)), UpHill(position=Position(1, 1)), UpHill(position=Position(1, 2))],
        [Valley(position=Position(2, 0)), Valley(position=Position(2, 1)), Valley(position=Position(2, 2))],
    ]
    start = grid[0][0]  # Top-left corner
    end = grid[2][2]  # Bottom-right corner

    # Expected path: left-bottom border path
    positions = [Position(1, 0), Position(2, 0), Position(2, 1), Position(2, 2)]
    expected_path = [Valley(position=position)  for position in positions]

    path = AStarStrategy.find_path(grid, start, end)

    assert path == expected_path


def test_dijkstra_lower_border_path():
    grid = [
        [Valley(position=Position(0, 0)), UpHill(position=Position(0, 1)), UpHill(position=Position(0, 2))],
        [Valley(position=Position(1, 0)), UpHill(position=Position(1, 1)), UpHill(position=Position(1, 2))],
        [Valley(position=Position(2, 0)), Valley(position=Position(2, 1)), Valley(position=Position(2, 2))],
    ]
    start = grid[0][0]  # Top-left corner
    end = grid[2][2]  # Bottom-right corner

    # Expected path: left-bottom border path, skipping bottom-left corner
    positions = [Position(1, 0), Position(2, 1), Position(2, 2)]
    expected_path = [Valley(position=position)  for position in positions]

    path = DijkstraStrategy.find_path(grid, start, end)

    assert path == expected_path
